import pandas as pd

"""
The feature data still contains some minor errors, such as non-uniform column names and missing country 2-digit codes, ... .
This file includes a method for each feature data source, which transforms the fetched data into the correct format.
These adjustments are only temporary, and the data should be corrected later (and the jobs updated accordingly), 
making this file obsolete in the future. (see issue #219)
"""


def historical_yield_data_renaming(data: pd.DataFrame) -> pd.DataFrame:

    # columns renaming
    renamings = {
        'valueType': 'ValueType',
        'year': 'Year',
        'crop': 'Crop',
        'spatialResolution': 'SpatialResolution',
        'country': 'Country',
        'state': 'State',
        'municipality': 'Municipality',
        'value': 'Value'
    }
    data = data.rename(columns=renamings)

    return data


def vi_data_renaming(data: pd.DataFrame) -> pd.DataFrame:

    # columns renaming
    renamings = {
        'Fpar': 'FPAR',
        'NAME': 'Country',
        'GID_0': 'CountryCode',
        'GEOID': 'CountryCode',
    }
    data = data.rename(columns=renamings)

    # the `CountryCode` colum sometimes contains three-digit country codes -> can simply cut to two-digit codes
    data['CountryCode'] = data['CountryCode'].str.slice(0, 2)

    # the US vi data contains an unnecessary column `AFFGEOID`
    data = data.drop(columns=['AFFGEOID'], errors='ignore')

    # remove unnecessary column `COUNTRY`
    data = data.drop(columns=['COUNTRY'], errors='ignore')

    return data


def climate_data_renaming(data: pd.DataFrame) -> pd.DataFrame:

    # columns renaming
    renamings = {
        'Country': 'CountryCode',
        'COUNTRY': 'Country',
        'NAME': 'Country',
        'GID_0': 'CountryCode',
        'GEOID': 'CountryCode',
    }
    data = data.rename(columns=renamings)

    # the climate data contains an unnecessary column `Unnamed: 0`
    data = data.drop(columns=['Unnamed: 0'], errors='ignore')

    return data


def forecast_data_renaming(data: pd.DataFrame) -> pd.DataFrame:

    # columns renaming
    renamings = {
        'country': 'Country',
        'state': 'State',
        'municipality': 'Municipality'
    }
    data = data.rename(columns=renamings)

    # create missing column containing '2 digits country code'
    country_to_country_code = {
        'Brazil': 'BR',
        'Argentina': 'AR',
        'Germany': 'DE',
        'United States': 'US'
    }
    data['CountryCode'] = data['Country'].apply(lambda c: country_to_country_code[c])

    return data


def soil_data_renaming(data: pd.DataFrame) -> pd.DataFrame:

    # columns renaming
    renamings = {
        'GID_0': 'CountryCode',
    }
    data = data.rename(columns=renamings)

    # the `CountryCode` colum sometimes contains three-digit country codes -> can simply cut to two-digit codes
    data['CountryCode'] = data['CountryCode'].str.slice(0, 2)

    # remove unnecessary column `COUNTRY`
    data = data.drop(columns=['COUNTRY'], errors='ignore')

    return data




from datetime import datetime
from io import StringIO
from typing import Optional

import pandas as pd
import requests
from requests.exceptions import HTTPError
from tqdm.notebook import tqdm

from .initialize import get_auth, get_fdc, is_initialized
from .utilities.feature_data_renamings import (climate_data_renaming, forecast_data_renaming,
                                               historical_yield_data_renaming, soil_data_renaming, vi_data_renaming)

GREENHUB_REST_API_URL: str = (
    "https://us-central1-digitalyieldmonitoringplatform.cloudfunctions.net"
)
SPATIAL_RESOLUTIONS: [str] = ["country", "state", "municipality"]
TIME_RESOLUTIONS: [str] = ["daily", "weekly", "monthly"]
SOIL_LAYERS: [str] = [f"D{i}" for i in range(1, 8)]
WARNING_COLOR_PRINT: str = "\033[93m"


def _fetch_data(file_path: str) -> Optional[pd.DataFrame]:
    """
    The greenhub REST API endpoint `getFeatureData` (Google Cloud Function) is queried with the given `file_path`.
    If the request is successful, the retrieved data is returned as a pandas DataFrame,
    if the REST API returns a status code 404 (Not Found), `None` is returned.

    :param file_path: paths to feature date in greenhub Google Cloud Storage `/featureData`
    :returns: dataframe containing the loaded data; None if res status code equals 404
    :raises HTTPError: if response status is not 200 or 404
    """

    # request data
    authentication = get_auth()
    params = {
        "featureDataPath": file_path,
    }
    auth_headers = authentication.get_auth_headers()
    res = requests.get(
        GREENHUB_REST_API_URL + "/getFeatureData", params=params, headers=auth_headers
    )

    # parse data
    if res.status_code == 200:
        csv_content = StringIO(res.text)
        df = pd.read_csv(csv_content)
        return df
    elif res.status_code == 404:
        return None
    else:
        raise HTTPError(
            f"Fetching feature data from API endpoint failed with status code: {res.status_code}"
        )


def _load_data(feature_files_paths: [str], data_name: str) -> Optional[pd.DataFrame]:
    """
    For each path in the `feature_files_paths`, it is first checked whether the data is already cached locally.
    If this is not the case, the data is loaded via the greenhub REST API `getFeatureData`.
    All loaded data is concatenated into a single pandas dataframe at the end,
    so it is important that the data referenced by the paths has the same structure (columns).

    :param feature_files_paths: paths to feature files in greenhub Google Cloud Storage `/featureData`
    :param data_name: the name of the data to load, used for logging purposes only
    :returns: dataframe containing the loaded and concatenated data, None if none of the paths worked
    """

    # fetch data and cache it, if already cached load it
    fdc = get_fdc()
    data = []
    for f in tqdm(feature_files_paths, desc=f"Loading {data_name}", unit="items"):
        cached_data = fdc.load_feature_data(f)
        if cached_data is None:
            # data not cached -> fetch data
            fetched_data = _fetch_data(f)
            if fetched_data is not None:
                data.append(fetched_data)
                fdc.save_feature_data(f, fetched_data)
        else:
            # data is cached
            data.append(cached_data)

    # combine all data to one single dataframe
    if not data:
        return None
    return pd.concat(data, axis=0, ignore_index=True)


def get_vi_data(
    country: str,
    start_year: int,
    end_year: int = None,
    spatial_resolution: str = "country",
) -> pd.DataFrame:
    """
    This function returns all available VI feature data on greenhub.ai for a specific `country` (e.g., US, BR)
    and year (`start_year`). By default, only VI data aggregated at the country level is returned,
    but with the `spatial_resolution parameter`, one can choose between 'country', 'state', and 'municipality'.
    Additionally, multiple years can be requested at once by specifying a time period by defining
    `start_year` **and** `end_year`.
    For more information please check out the official greenhub.ai documentation.

    :param country: country for which data is loaded, represented by a two-letter country code (e.g. US)
    :param start_year: year for which data is loaded; if `end_year` is also defined, then `start_year` is the
        first year of the time interval for which data is loaded
    :param end_year: if defined, `end_year` is the last year of the time interval, defined by `start_year` and
        `end_year`, for which data is loaded
    :param spatial_resolution: spatial resolution on which the data is aggregated; set to 'country' by default;
        can be set to 'country', 'state', or 'municipality'
    :returns: dataframe containing the loaded and concatenated data
    """

    # parameter None/default handling and preprocessing
    if end_year is None:
        end_year = start_year
    spatial_resolution = spatial_resolution.lower()
    country = country.upper()

    # check if greenhub api is initialized
    if not is_initialized():
        raise Exception("Please use greenhub.initialize() before requesting data.")

    # some obvious/important parameter checks
    if len(country) != 2:
        raise ValueError(
            "The country must be represented by a two-letter country code (e.g. US)."
        )
    if spatial_resolution not in SPATIAL_RESOLUTIONS:
        raise ValueError(
            f"The given spatial_resolution is not known. "
            f'These spatial_resolutions are available: {", ".join(SPATIAL_RESOLUTIONS)}.'
        )
    if start_year > end_year:
        raise ValueError(
            "The end_year must be greater than or equal to the start_year."
        )
    if start_year > datetime.now().year or end_year > datetime.now().year:
        raise ValueError(
            f"The start_year and end_year must be less than or equal to the current year."
        )
    if start_year < 2000 or end_year < 2000:
        print(
            f"Attention: There is no VI data available before 2000."
            f"\nFor more information please checkout the official greenhub.ai documentation. {WARNING_COLOR_PRINT}"
        )

    # construct all data files paths needed for the requested vi data
    feature_files_paths = []
    base_path = f"featureData/{country}/vi/{spatial_resolution}"
    for y in range(start_year, end_year + 1):
        for m in range(1, 13):
            feature_files_paths.append(f"{base_path}/{y}-{str(m).zfill(2)}.csv")

    # load and fetch data as one combined dataframe
    data = _load_data(feature_files_paths, "VI data")

    # check if data is empty -> meaning that not at least one path worked
    if data is None:
        raise ValueError(
            "For the given parameters, no data can be retrieved. "
            "For more information please checkout the official greenhub.ai documentation."
        )

    # inform user about missing years
    fetched_years = set(data["Year"].to_list())
    expected_years = set(range(start_year, end_year + 1))
    missing_years = expected_years - fetched_years
    if missing_years:
        print(
            f"Attention: no VI data found for {', '.join([str(i) for i in sorted(missing_years)])} {WARNING_COLOR_PRINT}"
        )

    return vi_data_renaming(data)


def get_climate_data(
    country: str,
    start_year: int,
    end_year: int = None,
    spatial_resolution: str = "country",
    time_resolution: str = "monthly",
) -> pd.DataFrame:
    """
    This function returns all available climate feature data on greenhub.ai for a specific `country` (e.g., US, BR)
    and year (`start_year`). By default, the climate data is aggregated at the country level and on a monthly basis,
    but with the `spatial_resolution parameter`, one can choose between 'country', 'state', and 'municipality'
    and the `time_resolution` parameter can be set to 'monthly', 'weekly', or 'daily'.
    Additionally, multiple years can be requested at once by specifying a time period by defining
    `start_year` **and** `end_year`.
    For more information please check out the official greenhub.ai documentation.

    :param country: country for which data is loaded, represented by a two-letter country code (e.g. US)
    :param start_year: year for which data is loaded; if `end_year` is also defined, then `start_year` is the
        first year of the time interval for which data is loaded
    :param end_year: if defined, `end_year` is the last year of the time interval, defined by `start_year` and
        `end_year`, for which data is loaded
    :param spatial_resolution: spatial resolution on which the data is aggregated; set to 'country' by default;
        can be set to 'country', 'state', or 'municipality'
    :param time_resolution: time resolution on which the data is aggregated; set to 'monthly' by default;
        can be set to 'monthly', 'weekly', or 'daily'
    :returns: dataframe containing the loaded and concatenated data
    """

    # parameter None/default handling and preprocessing
    if end_year is None:
        end_year = start_year
    spatial_resolution = spatial_resolution.lower()
    time_resolution = time_resolution.lower()

    # check if greenhub api is initialized
    if not is_initialized():
        raise Exception("Please use greenhub.initialize() before requesting data.")

    # some obvious/important parameter checks
    if len(country) != 2:
        raise ValueError(
            "The country must be represented by a two-letter country code (e.g. US)."
        )
    if start_year > end_year:
        raise ValueError(
            "The end_year must be greater than or equal to the start_year."
        )
    if spatial_resolution not in SPATIAL_RESOLUTIONS:
        raise ValueError(
            f"The given spatial_resolution is not known. "
            f'These spatial_resolutions are available: {", ".join(SPATIAL_RESOLUTIONS)}.'
        )
    if time_resolution not in TIME_RESOLUTIONS:
        raise ValueError(
            f"The given time_resolution is not known. "
            f'These time_resolution are available: {", ".join(TIME_RESOLUTIONS)}.'
        )
    if start_year > end_year:
        raise ValueError(
            "The end_year must be greater than or equal to the start_year."
        )
    if start_year > datetime.now().year or end_year > datetime.now().year:
        raise ValueError(
            f"The start_year and end_year must be less than or equal to the current year."
        )
    if start_year < 2000 or end_year < 2000:
        print(
            f"Attention: There is no climate data available before 2000."
            f"\nFor more information please checkout the official greenhub.ai documentation. {WARNING_COLOR_PRINT}"
        )

    # construct all data files paths needed for the requested climate data
    feature_files_paths = []
    base_path = (
        f"featureData/{country.upper()}/climate/{spatial_resolution}/{time_resolution}"
    )
    for y in range(start_year, end_year + 1):
        if time_resolution in ["monthly", "daily"]:
            for m in range(1, 13):
                feature_files_paths.append(f"{base_path}/{y}-{str(m).zfill(2)}.csv")
        if time_resolution == "weekly":
            for w in range(1, 54):
                feature_files_paths.append(
                    f"{base_path}/{y}-week-{str(w).zfill(2)}.csv"
                )

    # load and fetch data as one combined dataframe
    data = _load_data(feature_files_paths, "Climate data")

    # check if data is empty -> meaning that not at least one path worked
    if data is None:
        raise ValueError(
            "For the given parameters, no data can be retrieved. "
            "For more information please checkout the official greenhub.ai documentation."
        )

    # inform user about missing years
    fetched_years = set(data["Year"].to_list())
    expected_years = set(range(start_year, end_year + 1))
    missing_years = expected_years - fetched_years
    if missing_years:
        print(
            f"Attention: no Climate data found for {', '.join([str(i) for i in missing_years])} {WARNING_COLOR_PRINT}"
        )

    return climate_data_renaming(data)


def get_180d_forecast_data(
    country: str,
    start_year: int,
    start_month: int,
    spatial_resolution: str = "country",
    time_aggregation: str = "daily",
) -> pd.DataFrame:
    """
    This function returns the 180 days climate forecast data on greenhub.ai
    for a specific `country` (e.g., US, BR), year (`start_year`) and month (`start_month`).
    The prediction was made for the next 180 days starting from the selected start_year and start_month. For example,
    if one request data for January 2023, one will receive forecast data from month 1 (January) to month 6 (June),
    covering the next 180 days (6 months).

    By default, the forecast data is aggregated at country level and on a monthly basis,
    but with the `spatial_resolution` parameter, one can choose between 'country', 'state', and 'municipality'
    and the `time_aggregation` parameter can be set to 'monthly', 'weekly', or 'daily'.
    For more information please check out the official greenhub.ai documentation.

    :param country: country for which data is loaded, represented by a two-letter country code (e.g. US)
    :param start_year: start_year and start_month determine the point in time at which the 180-day prediction was made.
    :param start_month: start_year and start_month determine the point in time at which the 180-day prediction was made.
    :param spatial_resolution: spatial resolution on which the data is aggregated; set to 'country' by default;
        can be set to 'country', 'state', or 'municipality'
    :param time_aggregation: time resolution on which the data is aggregated; set to 'monthly' by default;
        can be set to 'monthly', 'weekly', or 'daily'
    :returns: dataframe containing the loaded and concatenated forecast data
    """

    # parameter None/default handling and preprocessing
    spatial_resolution = spatial_resolution.lower()
    time_aggregation = time_aggregation.lower()

    # check if greenhub api is initialized
    if not is_initialized():
        raise Exception("Please use greenhub.initialize() before requesting data.")

    # some obvious/important parameter checks
    if len(country) != 2:
        raise ValueError(
            "The country must be represented by a two-letter country code (e.g. US)."
        )
    if spatial_resolution not in SPATIAL_RESOLUTIONS:
        raise ValueError(
            f"The given spatial_resolution is not known. "
            f'These spatial_resolutions are available: {", ".join(SPATIAL_RESOLUTIONS)}.'
        )
    if time_aggregation not in TIME_RESOLUTIONS:
        raise ValueError(
            f"The given time_aggregation is not known. "
            f'These time_aggregations are available: {", ".join(TIME_RESOLUTIONS)}.'
        )
    if start_year < 2000:
        raise ValueError(
            f"There is no climate forecast data available before 2000. "
            f"For more information please checkout the official "
            f"greenhub.ai documentation. {WARNING_COLOR_PRINT}"
        )
    if start_month < 1 or start_month > 12:
        raise ValueError(
            f"The given start_month is not between 1 (january) and 12 (december)."
        )

    # construct all data files paths needed for the requested vi data
    feature_files_paths = [
        f"featureData/{country.upper()}/forecast/{spatial_resolution}/{time_aggregation}/"
        f"{start_year}-{str(start_month).zfill(2)}-180d.csv"
    ]

    # load and fetch data as one combined dataframe
    data = _load_data(feature_files_paths, "180d forecast data")

    # check if data is empty -> meaning that not at least one path worked
    if data is None:
        raise ValueError(
            "For the given parameters, no data can be retrieved. "
            "For more information please checkout the official greenhub.ai documentation."
        )

    return forecast_data_renaming(data)


def get_soil_data(
    country: str, spatial_resolution: str = "country", layer: str = None
) -> pd.DataFrame:
    """
    This function returns all available soil feature data on greenhub.ai for a specific `country` (e.g., US, BR)
    and year (`start_year`). By default, only VI data aggregated at the country level is returned,
    but with the `spatial_resolution parameter`, one can choose between 'country', 'state', and 'municipality'.
    Further, all 7 layers are returned by default. If one is only interested in one single layer, one can set the
    `layer` parameter to the specific layer, like for example 'D5'.
    The 7 layers 'D1' through 'D7' correspond with depth layers of 20 cm increments up until 100 cm ('D1': 0-20 cm,
    'D2': 20-40 cm, ..., 'D5': 80-100 cm), followed by layers of 50 cm increments until 200 cm ('D6': 100-150 cm,
    'D7': 150-200 cm).
    For more information please check out the official greenhub.ai documentation.

    :param country: country for which data is loaded, represented by a two-letter country code (e.g. US)
    :param spatial_resolution: spatial resolution on which the data is aggregated; set to 'country' by default;
        can be set to 'country', 'state', or 'municipality'
    :param layer: layer for which data is loaded; can be set to 'D1' up to 'D7'; set to None which loads all layers
    :returns: dataframe containing the loaded and concatenated data
    """

    # parameter None/default handling and preprocessing
    spatial_resolution = spatial_resolution.lower()
    layers = [layer.upper()] if layer is not None else SOIL_LAYERS

    # check if greenhub api is initialized
    if not is_initialized():
        raise Exception("Please use greenhub.initialize() before requesting data.")

    # some obvious/important parameter checks
    if len(country) != 2:
        raise ValueError(
            "The country must be represented by a two-letter country code (e.g. US)."
        )
    if any([l not in SOIL_LAYERS for l in layers]):
        raise ValueError(
            "The layer must be None (loading all layers) or set to 'D1' up to 'D7'."
        )

    # construct all data files paths needed for the requested vi data
    feature_files_paths = []
    base_path = f"featureData/{country.upper()}/soil/{spatial_resolution}"
    for l in layers:
        feature_files_paths.append(f"{base_path}/layer_{l}.csv")

    # load and fetch data as one combined dataframe
    data = _load_data(feature_files_paths, "Soil data")

    # check if data is empty -> meaning that not at least one path worked
    if data is None:
        raise ValueError(
            "For the given parameters, no data can be retrieved. "
            "For more information please checkout the official greenhub.ai documentation."
        )

    return soil_data_renaming(data)


def _get_spi_sti_data(
    country: str,
    spatial_resolution: str = "country",
    time_aggregation: str = "monthly",
    data_type: str = "spi",
) -> pd.DataFrame:
    """
    This function implements the redundant data fetching logic for `get_spi_data` and `get_sti_data`.
    It should not be accessed by the end-user. The end-user should only use `get_spi_data` and `get_sti_data`.
    """

    # parameter None/default handling and preprocessing
    spatial_resolution = spatial_resolution.lower()
    time_aggregation = time_aggregation.lower()

    # check if greenhub api is initialized
    if not is_initialized():
        raise Exception("Please use greenhub.initialize() before requesting data.")

    # some obvious/important parameter checks
    if len(country) != 2:
        raise ValueError(
            "The country must be represented by a two-letter country code (e.g. US)."
        )
    if spatial_resolution not in SPATIAL_RESOLUTIONS:
        raise ValueError(
            f"The given spatial_resolution is not known. "
            f'These spatial_resolutions are available: {", ".join(SPATIAL_RESOLUTIONS)}.'
        )
    if time_aggregation not in TIME_RESOLUTIONS[1:]:
        raise ValueError(
            f"The given time_aggregation is not known. "
            f'These time_aggregations are available: {", ".join(TIME_RESOLUTIONS[1:])}.'
        )

    # construct all data files paths needed for the requested data
    feature_files_paths = [
        f"featureData/{country.upper()}/distributionValues/{spatial_resolution}/{time_aggregation}/{data_type}.csv"
    ]

    # load and fetch data as one combined dataframe
    data = _load_data(feature_files_paths, f"{data_type.upper()} data")

    # check if data is empty -> meaning that not at least one path worked
    if data is None:
        raise ValueError(
            "For the given parameters, no data can be retrieved. "
            "For more information please checkout the official greenhub.ai documentation."
        )

    return data


def get_spi_data(
    country: str, spatial_resolution: str = "country", time_aggregation: str = "monthly"
) -> pd.DataFrame:
    """
    This function returns all available SPI data on greenhub.ai for a specific `country` (e.g., US, BR).
    By default, the SPI data is aggregated at country level and on a monthly basis,
    but with the `spatial_resolution` parameter, one can choose between 'country', 'state', and 'municipality'
    and the `time_aggregation` parameter can be set to 'monthly' or 'weekly'.
    For more information please check out the official greenhub.ai documentation.

    :param country: country for which data is loaded, represented by a two-letter country code (e.g. US)
    :param spatial_resolution: spatial resolution on which the data is aggregated; set to 'country' by default;
        can be set to 'country', 'state', or 'municipality'
    :param time_aggregation: time resolution on which the data is aggregated; set to 'monthly' by default;
        can be set to 'monthly', 'weekly'
    :returns: dataframe containing the loaded and concatenated SPI data
    """
    return _get_spi_sti_data(country, spatial_resolution, time_aggregation, "spi")


def get_sti_data(
    country: str, spatial_resolution: str = "country", time_aggregation: str = "monthly"
) -> pd.DataFrame:
    """
    This function returns all available STI data on greenhub.ai for a specific `country` (e.g., US, BR).
    By default, the STI data is aggregated at country level and on a monthly basis,
    but with the `spatial_resolution` parameter, one can choose between 'country', 'state', and 'municipality'
    and the `time_aggregation` parameter can be set to 'monthly' or 'weekly'.
    For more information please check out the official greenhub.ai documentation.

    :param country: country for which data is loaded, represented by a two-letter country code (e.g. US)
    :param spatial_resolution: spatial resolution on which the data is aggregated; set to 'country' by default;
        can be set to 'country', 'state', or 'municipality'
    :param time_aggregation: time resolution on which the data is aggregated; set to 'monthly' by default;
        can be set to 'monthly', 'weekly'
    :returns: dataframe containing the loaded and concatenated STI data
    """
    return _get_spi_sti_data(country, spatial_resolution, time_aggregation, "sti")


def get_historical_yield_data(
    crop: str,
    year: int,
    country: str,
    spatial_resolution: str = "country",
) -> pd.DataFrame:
    """
    This function returns all available historical yield data on greenhub.ai for a specific crop
    (e.g.`Soybeans`, `Wheat-Winter`, `Wheat`), year (since 2000), `country` (e.g., US, BR).
    By default, only historical yield data at the country level is returned,
    but with the `spatial_resolution` parameter, one can choose between 'country', 'state', and 'municipality'.
    For more information please check out the official greenhub.ai documentation.

    :param crop: crop for which data is loaded;
    :param year: year for which data is loaded;
    :param country: country for which data is loaded, represented by a two-letter country code (e.g. US)
    :param spatial_resolution: spatial resolution on which the data is aggregated; set to 'country' by default;
        can be set to 'country', 'state', or 'municipality'
    :returns: dataframe containing the loaded and concatenated data
    """

    # parameter None/default handling and preprocessing
    crop = crop.lower()
    country = country.upper()
    spatial_resolution = spatial_resolution.lower()

    # check if greenhub api is initialized
    if not is_initialized():
        raise Exception("Please use greenhub.initialize() before requesting data.")

    # some obvious/important parameter checks
    if len(country) != 2:
        raise ValueError(
            "The country must be represented by a two-letter country code (e.g. US)."
        )
    if spatial_resolution not in SPATIAL_RESOLUTIONS:
        raise ValueError(
            f"The given spatial_resolution is not known. "
            f'These spatial_resolutions are available: {", ".join(SPATIAL_RESOLUTIONS)}.'
        )
    if year > datetime.now().year:
        raise ValueError(f"The year must be less or equal than the current year.")
    if year < 2000:
        print(
            f"Attention: There is no historical yield data available before 2000."
            f"\nFor more information please checkout the official greenhub.ai documentation. {WARNING_COLOR_PRINT}"
        )

    # construct data file path needed for the requested historical yield data
    feature_files_paths = [
        f"historicalData/{country}/{crop}/yield/{year}/{spatial_resolution}"
    ]

    # load and fetch data as one combined dataframe
    data = _load_data(feature_files_paths, "Historical Yield data")

    # check if data is empty -> meaning that not at least one path worked
    if data is None:
        raise ValueError(
            "For the given parameters, no data can be retrieved. "
            "For more information please checkout the official greenhub.ai documentation."
        )

    return historical_yield_data_renaming(data)



from typing import Optional
from .utilities.auth import get_authentication, Authentication
from .utilities.feature_data_cache import FeatureDataCache


authentication: Optional[Authentication] = None
fdc: Optional[FeatureDataCache] = None


def initialize(api_key: str):
    """
    Initialize the greenhub SDK.
    It is important to initialize the greenhub SDK before using any further functionalities of the SDK.
    The given api_key is passed on and checked by the authentication module.

    Note: The method is cloud aware and ignores the api_key in case the method is used in scripts
    running in the gcloud.

    :param api_key: user API key derived from greenhub.ai
    """

    global authentication
    global fdc

    authentication = get_authentication(api_key)
    fdc = FeatureDataCache()


def is_initialized() -> bool:
    return (not fdc is None) and (not authentication is None)


def get_auth() -> Optional[Authentication]:
    global authentication
    return authentication


def get_fdc() -> Optional[FeatureDataCache]:
    global fdc
    return fdc



import os
import shutil
import tempfile
import pandas as pd
from typing import Optional


TEMP_DIR_DEFAULT_NAME = 'greenhub_feature_data_temp_8410_'


class FeatureDataCache:
    """
    Functionalities that need to repeatedly load feature data from an API can cache this data
    using a `FeatureDataCache` object, thereby avoiding unnecessary reloading.
    The stored data is saved in the background in a dedicated temp directory and remains available even after closing
    the terminal, environment, etc. The cached data is only lost upon restarting the computer.
    """

    def __init__(self):
        """
        Initializes the `FeatureDataCache` instance by setting up the necessary dedicated temp directory,
        where the given data will be cached as csv files.

        Special/Workaround: `tempfile.mkdtemp(...)` creates a temp directory with a specific prefix
        and a random code as the name, ensuring the folder is unique. However, since we want to access
        the data even after the program ends and have no way to securely store this name, we first search
        in the computer's `temp` directory for a directory starting with our `TEMP_DIR_DEFAULT_NAME` prefix.
        """

        self.system_temp_dir_path = tempfile.gettempdir()

        # get greenhub temp dir path, create if not already exists
        self.greenhub_temp_dir_path = None
        for item in os.listdir(self.system_temp_dir_path):
            item_path = f'{self.system_temp_dir_path}/{item}'
            if os.path.isdir(item_path) and item.startswith(TEMP_DIR_DEFAULT_NAME):
                # greenhub temp dir already exists
                self.greenhub_temp_dir_path = item_path
                break
        if self.greenhub_temp_dir_path is None:
            # create new greenhub temp dir
            self.greenhub_temp_dir_path = tempfile.mkdtemp(prefix=TEMP_DIR_DEFAULT_NAME)

    def save_feature_data(self, feature_file_path: str, data: pd.DataFrame):
        """
        Caches the given (feature) `data` as a csv file in the dedicated temp directory.

        :param feature_file_path: the relative, unique path within the temp directory where the data should be stored.
        :data: the (feature) data to be cached.
        """

        file_name = os.path.basename(feature_file_path)
        dir_path = os.path.join(
            self.greenhub_temp_dir_path,
            os.path.dirname(feature_file_path)
        )
        os.makedirs(dir_path, exist_ok=True)
        data.to_csv(os.path.join(dir_path, file_name), index=False)

    def load_feature_data(self, feature_file_path: str) -> Optional[pd.DataFrame]:
        """
        Loads cached (feature) `data` and returns is as a pandas dataframe.
        The stored data is uniquely identified by the passed `feature_file_path`.

        :param feature_file_path: the relative path within the temp directory which uniquely identified the stored data.
        :returns: the loaded data; None is returned if no data is found.
        """

        feature_file_path = os.path.join(
            self.greenhub_temp_dir_path,
            feature_file_path
        )
        if not os.path.exists(feature_file_path):
            return None
        return pd.read_csv(feature_file_path)

    def reset(self):
        """
        Reset entire cache. All cached data is deleted.
        """
        shutil.rmtree(self.greenhub_temp_dir_path)
        self.__init__()




import os
import sys
import inspect
import importlib


def greenhub_cli_script():
    """
    This method is registered as an Entry Point (console script) during the package installation,
    allowing the user to execute the command `greenhub <command>` in the terminal
    (more details: https://setuptools.pypa.io/en/latest/userguide/entry_point.html).
    Here, the provided `<command>` is parsed, and the corresponding functionality to be executed is called.
    """

    request_action = sys.argv[1]
    if request_action in ['help', '-h', '--help']:
        print_cli_description()
        return
    if request_action == 'test':
        run_implementation_test()
        return

    print(f"Error: given command `{request_action}` is not known by greenhub. "
          f"Please checkout the official greenhub.ai documentation.")


def print_cli_description():
    """
    Print a textual description of the `greenhub` command-line interface.
    """

    sections = []
    sections.append(
        "Welcome to the command-line interface of the `greenhub` package."
        "\nThe following describes the available command-line functionalities. "
        "\nFor more information, please visit docs.greenhub.ai."
    )
    sections.append(
        "Usage:"
        "\n  greenhub <command>"
    )
    sections.append(
        "Commands:"
        "\n  help     Show help for commands."
        "\n  test     Test if the current directory meets all requirements to be uploaded "
        "as a zip file (model) to greenhub.ai."
    )
    print('\n'.join(sections))


def run_implementation_test():
    """
    Test if the current directory meets all requirements to be uploaded as a zip file (model) to greenhub.ai.
    For now, it is only tested if the current directory contains a `run.py` file with a `run` with a specific signature.
    """

    RUN_FILE_NAME = 'run.py'
    RUN_FUNCTION_NAME = 'run'

    # check if a 'run.py' file exists in the current directory
    run_file_path = os.path.join(os.getcwd(), RUN_FILE_NAME)
    if not os.path.exists(run_file_path) or not os.path.isfile(run_file_path):
        print("FAILURE: no 'run.py' file found in the current directory. "
              "Please checkout the official greenhub.ai documentation.")
        return

    # check 'run.py' file for 'run' function implementation -> load function if exists
    # load 'run.py' module
    spec = importlib.util.spec_from_file_location(RUN_FILE_NAME[:-3], run_file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    # check if 'run' exists and load it
    if hasattr(module, 'run'):
        run_func = getattr(module, RUN_FUNCTION_NAME)
    else:
        print("FAILURE: found no 'run' function implementation in 'run.py'. "
              "For more information please checkout the official greenhub.ai documentation.")
        return

    # check 'run' function parameter names and types
    run_signature = inspect.signature(run_func)
    params = run_signature.parameters.values()
    param_names = [p.name for p in params]
    param_types = [p.annotation.__name__ for p in params]
    # expected signatures
    EXPECTED_SIGNATURE = {
        'names': ['year', 'month'],
        'types': ['int', 'int'],
        'signature': 'run(year: int, month: int)'
    }
    # check
    if param_names != EXPECTED_SIGNATURE['names']:
        print(f"FAILURE: 'run' function implementation does not expect the right parameters. "
              f"The 'run' function signature should look like: '{EXPECTED_SIGNATURE['signature']}' "
              f"For more information please checkout the official greenhub.ai documentation.")
        return
    if param_types != EXPECTED_SIGNATURE['types']:
        print(f"FAILURE: 'run' function implementation required parameters are of the wrong type. "
              f"The 'run' function signature should look like: '{EXPECTED_SIGNATURE['signature']}' "
              f"For more information please checkout the official greenhub.ai documentation.")
        return

    # success
    print("SUCCESS: the 'run' implementation was checked successfully and can now be uploaded to greenhub.ai. "
          "For more information and step by step tutorials, please checkout greenhub.ai.")



import greenhub as gh
import pandas as pd


def run(year: int, month: int):
    
    # Initialize GreenHub SDK
    # TODO: You can find your API key in your Account dialog at the top right of the GreenHub page
    gh.initialize("YOUR_API_KEY")

    # Fetch and setup feature vector
    features = ...  # TODO

    # Load model
    model = ...  # TODO

    # Run model
    prediction = ...  # TODO

    # Format to expected GreenHub output
    output = ...  # TODO

    return output



import greenhub as gh
import pandas as pd
import random


def run(year: int, month: int):

    # Fetch and setup feature vector
    states = ['WEST VIRGINIA', 'WASHINGTON', 'TENNESSEE', 'MISSOURI']

    # Load model
    avg_values = [3.0, 4.0, 3.0, 4.0]
    year_factor = (year - 2000) * 0.01
    month_factor = month * 0.05

    # Run model
    predictions = [a + year_factor + month_factor + random.uniform(-0.4, 0.6) for a in avg_values]

    # Format to expected GreenHub output
    output_dict = {'state': [states], 'prediction': predictions}
    output = pd.DataFrame(output_dict)

    return output



from .data import get_vi_data, get_climate_data, get_180d_forecast_data, get_soil_data, get_sti_data, get_spi_data
from .initialize import initialize

__all__ = [
    'initialize',

    'get_vi_data',
    'get_climate_data',
    'get_soil_data',
    'get_180d_forecast_data'
]



import greenhub as gh
import pandas as pd
import pickle


def run(year: int, month: int):
    
    # Set  run parameters
    country = 'AR'
    state = 'SANTA FE'
    spatial_resolution = 'state'
    
    # Initialize GreenHub SDK (note that this step will be removed as soon as in production)
    gh.initialize("-")
    
    # Create feature vector
    features = create_feature_vector(country, year, spatial_resolution, state)

    # Load model
    with open('linear_regression.pkl', 'rb') as f:
        model = pickle.load(f)

    # Run model
    prediction = model.predict(features)

    # Format to expected GreenHub output
    output_dict = {'state': [state], 'prediction': prediction.tolist()}
    output = pd.DataFrame(output_dict)

    # Return model prediction
    return output


def create_feature_vector(country, start_year, spatial_resolution, state):

    # Prepare VI data
    vi = gh.get_vi_data(country=country, start_year=start_year, spatial_resolution=spatial_resolution)
    vi = vi[['EVI', 'NDVI', 'State', 'Year', 'Month']]
    vi['Year'] = pd.to_numeric(vi['Year'])
    vi = vi[vi['State'] == state]

    # Prepare soil data
    soil = gh.get_soil_data(country=country, spatial_resolution=spatial_resolution, layer='D1')
    selected_cols = [col for col in soil.columns if
                     (col.endswith('_avg') and not col.startswith('TP-')) or col in ['Layer', 'NAME_1']]
    soil = soil[selected_cols]
    soil.columns = soil.columns.str.replace('_avg', '')
    soil.rename({'NAME_1': 'State'}, axis=1, inplace=True)
    soil['State'] = soil['State'].apply(lambda x: x.upper())
    
    # Prepare climate data
    climate = gh.get_climate_data(country='AR', start_year=2010, end_year=2023, spatial_resolution='state')
    climate = climate.drop(columns=['CountryCode'])
    climate['Year'] = pd.to_numeric(climate['Year'])
    
    # Merge data
    merged_df = pd.merge(vi, soil, on='State', how='left')
    merged_df = pd.merge(merged_df, climate, on=['Year', 'Month', 'State'], how='left')
    merged_df = merged_df.dropna()

    # Create a pivot table
    pivot_df = merged_df.pivot_table(index=['Year', 'State'], columns='Month', aggfunc='first')

    # Flatten the MultiIndex columns
    pivot_df.columns = ['{}_Month{}'.format(col[0], int(col[1])) for col in pivot_df.columns]

    # Reset index to turn MultiIndex into columns
    pivot_df.reset_index(inplace=True)

    # Fix 12x columns for static values
    squash_cols = ['Drain', 'CFRAG', 'SDTO', 'STPC', 'CLPC', 'BULK', 'TAWC', 'CECS', 'BSAT', 'ESP',
                   'CECc', 'PHAQ', 'TCEQ', 'GYPS', 'ELCO', 'ORGC', 'TOTN', 'CNrt', 'ECEC', 'ALSA', 'Value']
    pivot_df = pivot_df.rename({f'{squash}_Month1': f'{squash}' for squash in squash_cols}, axis=1)
    pivot_df = pivot_df.drop(
        columns=[col for col in pivot_df.columns if any(col.startswith(f'{squash}_Month') for squash in squash_cols)])

    # Limit the feature vector to state
    single_state_df = pivot_df[pivot_df['State'] == state.upper()]
    single_state_df.head()

    # Final feature vector
    feature_vector = single_state_df.drop(columns=['Year', 'State'])
    feature_vector = feature_vector.dropna()

    return feature_vector



import os
from urllib.error import HTTPError
import requests

API_KEY_CHECK_REST_API_URL: str = 'https://us-central1-digitalyieldmonitoringplatform.cloudfunctions.net/checkApiKey'
ID_TOKEN_AUDIENCE: str = "https://us-central1-digitalyieldmonitoringplatform.cloudfunctions.net/"


class Authentication:

    def get_auth_headers(self) -> dict:
        raise NotImplementedError()


class ApiKeyAuthentication(Authentication):

    def __init__(self, api_key: str):
        self.api_key = api_key
        res = requests.get(API_KEY_CHECK_REST_API_URL, params={'apiKey': self.api_key})

        if res.status_code == 401:
            raise ValueError("Invalid API key.")
        if res.status_code != 200:
            raise HTTPError(f'Checking API key failed with status code: {res.status_code}')

    def get_auth_headers(self) -> dict:
        return {'api_key': self.api_key}


class GCloudIDTokenAuthentication(Authentication):

    def __init__(self):
        # Define the metadata URL
        url = f"http://metadata.google.internal/computeMetadata/v1/instance/service-accounts/default/identity?audience={ID_TOKEN_AUDIENCE}"

        # Set the headers
        headers = {
            "Metadata-Flavor": "Google"
        }

        # Make the GET request
        response = requests.get(url, headers=headers)

        if response.status_code != 200:
            raise HTTPError(f"Failed to fetch identity: {response.status_code}, {response.text}")

        self.id_token = response.text

    def get_auth_headers(self) -> dict:
        return {'Authorization': f'Bearer {self.id_token}'}


def is_running_in_cloud() -> bool:
    return os.getenv('CLOUD_RUN_JOB') is not None


def get_authentication(api_key: str) -> Authentication:
    if is_running_in_cloud():
        print("Using cloud auth")
        return GCloudIDTokenAuthentication()
    else:
        print("Using api key auth")
        return ApiKeyAuthentication(api_key)
